import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange
from einops.layers.torch import Rearrange
from timm.models.layers import DropPath, trunc_normal_

from enum import Enum, auto


class AttentionType(Enum):
    LOCAL = auto()
    GLOBAL = auto()


class PositionalEncodingType(Enum):
    RELATIVE = auto()
    ABSOLUTE = auto()


class MaybeDropout(nn.Module):
    def __init__(self, dropout_rate: float):
        super().__init__()

        if dropout_rate > 0.0:
            self.func = nn.Dropout(p=dropout_rate)
        else:
            self.func = nn.Identity()

    def forward(self, x):
        return self.func(x)


class WindowedAttention(nn.Module):
    def __init__(self,
                 attention_type: AttentionType,
                 positional_encoding_type: PositionalEncodingType,
                 window_size: int,
                 input_size: int,
                 feature_size: int,
                 num_heads: int):
        super().__init__()

        assert feature_size >= num_heads and feature_size % num_heads == 0, "Embedding size is incompatible with number of heads!"
        assert input_size >= window_size and input_size % window_size == 0, "Window size is incompatible with input size!"

        self.attention_type = attention_type
        self.positional_encoding_type = positional_encoding_type
        self.window_size = window_size
        self.input_size = input_size
        self.feature_size = feature_size
        self.scale = self.feature_size ** (-0.5)
        self.num_heads = num_heads

        self.num_windows_per_axis = self.input_size // self.window_size

        self.qkv_layer = nn.Linear(self.feature_size, 3 * self.feature_size)
        if self.positional_encoding_type is PositionalEncodingType.RELATIVE:
            self.relative_position_bias_table = nn.Parameter(
                torch.zeros((2 * self.window_size - 1) ** 3),
                requires_grad=True
            )
            trunc_normal_(self.relative_position_bias_table, std=.02)

            # Compute pairwise differences between 3d coordinates
            ix = torch.arange(self.window_size)  # [WS]
            coords = torch.stack(torch.meshgrid([ix, ix, ix], indexing='ij'), dim=0)  # [3, WS, WS, WS]
            coords_flatten = torch.flatten(coords, start_dim=1)  # [3, WS^3]
            relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # [3, WS^3, WS^3]
            relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # [WS^3, WS^3, 3]

            # Shift to interval [0, 2 * (WS - 1)]
            relative_coords += self.window_size - 1

            # 3d indexing to linear indexing
            relative_coords[:, :, 0] *= (2 * self.window_size - 1) ** 2
            relative_coords[:, :, 1] *= 2 * self.window_size - 1
            relative_position_index = relative_coords.sum(-1)  # [WS^3, WS^3]

            self.register_buffer("relative_position_index", relative_position_index)
        elif self.positional_encoding_type is PositionalEncodingType.ABSOLUTE:
            self.absolute_position_bias = nn.Parameter(torch.zeros(self.window_size ** 3, self.window_size ** 3),
                                                       requires_grad=True)
            trunc_normal_(self.absolute_position_bias, std=.02)
        else:
            raise NotImplementedError("Unknown positional encoding!")

        self.attention_softmax = nn.Softmax(dim=-1)

        self.head_proj = nn.Sequential(nn.LayerNorm(self.feature_size),
                                       nn.Linear(self.feature_size, self.feature_size))

    def forward(self,
                x: torch.FloatTensor  # [B, ND, NH, NW, E]
                ):

        if self.attention_type is AttentionType.GLOBAL:
            x = rearrange(x, ' b (kd s1) (kh s2) (kw s3) e -> b (s1 kd) (s2 kh) (s3 kw) e',
                          s1=self.num_windows_per_axis, s2=self.num_windows_per_axis, s3=self.num_windows_per_axis)

        x = rearrange(x, 'b (kd s1) (kh s2) (kw s3) e -> b (kd kh kw) (s1 s2 s3) e',
                      s1=self.window_size, s2=self.window_size, s3=self.window_size)  # [B, KD * KH * KW, WS^3, E]

        qkv = self.qkv_layer(x)  # [B, KD * KH * KW, WS^3, 3 * E]
        q, k, v = rearrange(qkv, 'b k w (l h f) -> l b k h w f', l=3, h=self.num_heads).chunk(3, dim=0)  # each [1, B, KD * KH * KW, #Heads, WS^3, F]
        q, k, v = q[0], k[0], v[0]  # each [B, KD * KH * KW, WS^3, F]
        qk_dot = self.scale * torch.einsum('ijkut,ijkvt->ijkuv', q, k)  # [B, KD * KH * KW, #Heads, WS^3, WS^3]

        if self.positional_encoding_type is PositionalEncodingType.RELATIVE:
            position_bias = self.relative_position_bias_table[self.relative_position_index]  # [WS^3, WS^3]
        elif self.positional_encoding_type is PositionalEncodingType.ABSOLUTE:
            position_bias = self.absolute_position_bias  # [WS^3, WS^3]
        else:
            raise NotImplementedError("Unknown positional encoding!")

        qk_dot_rel = qk_dot + position_bias  # [B, KD * KH * KW, #Heads, WS^3, WS^3]
        qk_dot_rel = self.attention_softmax(qk_dot_rel)

        x = torch.einsum('ijkut,ijktv->ijkuv', qk_dot_rel, v)  # [B, KD * KH * KW, #Heads, WS^3, F]
        x = rearrange(x, 'b k h w f -> b k w (h f)')  # [B, KD * KH * KW, WS^3, (#Heads * F = E)]
        x = self.head_proj(x)  # [B, KD * KH * KW, WS^3, E]

        x = rearrange(x, 'b (kd kh kw) (s1 s2 s3) e -> b (kd s1) (kh s2) (kw s3) e',
                      kd=self.num_windows_per_axis, kh=self.num_windows_per_axis, kw=self.num_windows_per_axis,
                      s1=self.window_size, s2=self.window_size, s3=self.window_size)  # [B, ND, NH, NW, E]

        if self.attention_type is AttentionType.GLOBAL:
            x = rearrange(x, 'b (s1 kd) (s2 kh) (s3 kw) e -> b (kd s1) (kh s2) (kw s3) e',
                          s1=self.window_size, s2=self.window_size, s3=self.window_size)  # [B, ND, NH, NW, E]

        return x


class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        return x + self.fn(x, **kwargs)


class TransformerBlock(nn.Module):
    def __init__(self,
                 attention_type: AttentionType,
                 positional_encoding_type: PositionalEncodingType,
                 embedding_size: int,
                 input_size: int,
                 window_size: int,
                 num_heads: int,
                 dropout_prob: float):
        super().__init__()

        self.attention_type = attention_type
        self.positional_encoding_type = positional_encoding_type
        self.embedding_size = embedding_size
        self.input_size = input_size
        self.window_size = window_size
        self.num_heads = num_heads
        self.dropout_prob = dropout_prob

        self.feature_attention_layer = nn.Sequential(nn.LayerNorm(self.embedding_size),
                                                     MaybeDropout(self.dropout_prob),
                                                     WindowedAttention(attention_type=self.attention_type,
                                                                       positional_encoding_type=self.positional_encoding_type,
                                                                       input_size=self.input_size,
                                                                       window_size=self.window_size,
                                                                       feature_size=self.embedding_size,
                                                                       num_heads=self.num_heads))
        self.feature_mlp = nn.Sequential(
            nn.LayerNorm(self.embedding_size),
            MaybeDropout(self.dropout_prob),
            nn.Linear(self.embedding_size, 4 * self.embedding_size),
            nn.GELU(),
            nn.LayerNorm(4 * self.embedding_size),
            MaybeDropout(self.dropout_prob),
            nn.Linear(4 * self.embedding_size, self.embedding_size)
        )

    def forward(self,
                x: torch.FloatTensor  # [B, ND, NH, NW, E]
                ):
        x = self.feature_attention_layer(x)  # [B, ND, NH, NW, E]
        x = self.feature_mlp(x)  # [B, ND, NH, NW, E]
        return x


class PatchEmbeddingLayer(nn.Sequential):
    def __init__(self,
                 patch_size: int,
                 embedding_size: int):
        super().__init__(
            Rearrange('b c (nd s1) (nh s2) (nw s3) -> b nd nh nw (s1 s2 s3 c)',
                      s1=patch_size,
                      s2=patch_size,
                      s3=patch_size),
            nn.Linear(patch_size ** 3, embedding_size)
        )


class ReversePatchEmbeddingLayer(nn.Sequential):
    def __init__(self,
                 num_patches_per_axis: int,
                 patch_size: int,
                 embedding_size: int
                 ):
        super().__init__(
            nn.LayerNorm(embedding_size),
            nn.Linear(embedding_size, patch_size ** 3),
            Rearrange('b nd nh nw (s1 s2 s3 c) -> b c (nd s1) (nh s2) (nw s3)',
                      nh=num_patches_per_axis, nw=num_patches_per_axis, nd=num_patches_per_axis,
                      s1=patch_size, s2=patch_size, s3=patch_size)
        )


class WG2LART(nn.Module):
    def __init__(self,
                 input_size: int,
                 patch_size: int,
                 embedding_size: int,
                 window_size: int,
                 num_heads: int,
                 num_blocks: int,
                 dropout_prob: float,
                 positional_encoding_type: str,
                 **kwargs):
        super().__init__()

        self.input_size = input_size
        self.patch_size = patch_size
        self.embedding_size = embedding_size
        self.window_size = window_size
        self.num_heads = num_heads
        self.num_blocks = num_blocks
        self.dropout_prob = dropout_prob
        self.positional_encoding_type = PositionalEncodingType[positional_encoding_type]

        self.num_patches_per_axis = self.input_size // self.patch_size

        self.patch_embedding = PatchEmbeddingLayer(
            patch_size=self.patch_size,
            embedding_size=self.embedding_size
        )

        self.reverse_patch_embedding = ReversePatchEmbeddingLayer(
            num_patches_per_axis=self.num_patches_per_axis,
            patch_size=self.patch_size,
            embedding_size=self.embedding_size
        )

        self.blocks = nn.Sequential(*[Residual(TransformerBlock(
            attention_type=[AttentionType.LOCAL, AttentionType.GLOBAL][i % 2],
            embedding_size=self.embedding_size,
            input_size=self.input_size // self.patch_size,
            window_size=self.window_size,
            num_heads=self.num_heads,
            dropout_prob=self.dropout_prob,
            positional_encoding_type=self.positional_encoding_type
        )) for i in range(self.num_blocks)])

    def forward(self,
                x: torch.FloatTensor  # [B, 1, D, H, W]
                ):
        y = self.patch_embedding(x)  # [B, ND, NH, NW, E]
        y = self.blocks(y)  # [B, ND, NH, NW, E]
        y = self.reverse_patch_embedding(y)  # [B, 1, D, H, W]

        return y

